package  belka.geom;

//--- Java imports ---
import java.io.*;

//--- Application imports ---
import belka.mol.*;
import Jama.*;

/**
 * The class calculates displacement vector between two given confomartions
 * of a protein. Protein sequences are supposed to be aligned. Rigid blocks
 * are supposed to be assigned.
 * 
 * @author Alexej Abyzov
 */
public class Displacer
{
    // Chains
    Chain chain1_ = null, chain2_ = null;
    Pair   fPair_ = null, lPair_  = null;

    // Local rotation matrices and translation vectors, i.e. for 
    // superimposing each rigid block. Element zero stores transformation for
    // global superposition.
    Matrix[] rotations_    = null;
    Matrix[] translations_ = null;

    // Array with indexes for relative movements
    int[] relative_index_ = null;
    
    /**
     * Object constructor. 
     *
     * @param chain1 chain with first conformation.
     * @param chain2 chain with second conformation.
     */
    public Displacer(Chain chain1,Chain chain2)
    {
	if (chain1 == null || chain2 == null) return;

	short nAligned = 0;
	for (Assembly a1 = chain1.assemblyList(), a2 = chain2.assemblyList();
	     a1 != null && a2 != null;a1 = a1.next(), a2 = a2.next()) {

	    boolean isGap = (a1.isGap() || a2.isGap());
	    Pair newPair = new Pair(a1,a2,!isGap);
	    if (a1.getGroupId() == a2.getGroupId())
		newPair.setIntValue(a1.getGroupId());

	    if (fPair_ == null) fPair_ = lPair_ = newPair;
	    else if (lPair_.insertAfter(newPair)) {
		Assembly a1to1 = (Assembly)lPair_.getObject1();
		Assembly a2to2 = (Assembly)lPair_.getObject2();
		if (a1.isConnectedTo(a1to1) && a2.isConnectedTo(a2to2)) {
		    lPair_.addConnection(newPair);
		    newPair.addConnection(lPair_);
		}
		lPair_ = newPair;
	    }
	    if (!isGap) nAligned++;
	}

	if (nAligned < 3) {
	    System.err.println("Not enough aligned atoms in chains.");
	    return;
	}

	// Storing chains
	chain1_ = chain1;
	chain2_ = chain2;

	// Calculating number of rigid blocks
	int n_rigids = 0;
	for (Pair p = fPair_;p != null;p = p.next()) {
	    if (!p.isOfInterest()) continue;
	    if (p.getIntValue() > n_rigids) n_rigids = p.getIntValue();
	}

	// Allocating array for matrices
	rotations_      = new Matrix[n_rigids + 1];
	translations_   = new Matrix[n_rigids + 1];
	relative_index_ = new int[n_rigids + 1];

	// Rotation and translation for global superposition
	Atom[] atoms1 = new Atom[nAligned];
	Atom[] atoms2 = new Atom[nAligned];
	int index = 0;
	for (Pair p = fPair_;p != null;p = p.next()) {
	    if (!p.isOfInterest()) continue;
	    Assembly ass1 = (Assembly)p.getObject1();
	    Assembly ass2 = (Assembly)p.getObject2();
	    Atom a1 = ass1.getMainAtom();
	    Atom a2 = ass2.getMainAtom();
	    if (a1 == null || a2 == null) continue;
	    atoms1[index] = a1;
	    atoms2[index] = a2;
	    index++;
	}

	Kabscher kb = new Kabscher();
	kb.fit(atoms1,atoms2);
	rotations_[0]    = new Matrix(kb.getRotation());
	translations_[0] = new Matrix(kb.getTranslation(),
				      kb.getTranslation().length);

	// Rotation and translation for local superpositions
	for (int r = 1;r <= n_rigids;r++) {
	    index = 0;
	    for (Pair p = fPair_;p != null;p = p.next()) {
		if (!p.isOfInterest()) continue;
		if (p.getIntValue() != r) continue;
		Assembly ass1 = (Assembly)p.getObject1();
		Assembly ass2 = (Assembly)p.getObject2();
		Atom a1 = ass1.getMainAtom();
		Atom a2 = ass2.getMainAtom();
		if (a1 == null || a2 == null) continue;
		atoms1[index] = a1;
		atoms2[index] = a2;
		index++;
	    }
	    if (nAligned < 3) {
		System.err.println("Not enough aligned atoms in rigid #" +
				   r + ".");
		continue;
	    }
	    while (index < nAligned) { // Clearing remaining cells
		atoms1[index] = atoms2[index] = null;
		index++;
	    }
	    kb = new Kabscher();
	    kb.fit(atoms1,atoms2);
	    rotations_[r]    = new Matrix(kb.getRotationCopy());
	    translations_[r] = new Matrix(kb.getTranslationCopy(),
					  kb.getTranslationCopy().length);
	}
    }

    /**
     * Returns string with values of displacement for each CA atom. The 
     * displacement is the vector between initial and final comformations.
     *
     * @return string with values of displacement for each CA atom.
     */
    public StringBuffer printSimple()
    {
	if (rotations_ == null || translations_ == null)
	    return new StringBuffer();

	Matrix rot   = rotations_[0];
	Matrix trans = translations_[0];
	if (rot == null || trans == null) {
	    System.err.println("No global tranformation found.");
	    return new StringBuffer();
	}
	
	// Calculating displacements
	double r00 = rot.get(0,0),r01 = rot.get(0,1),r02 = rot.get(0,2);
	double r10 = rot.get(1,0),r11 = rot.get(1,1),r12 = rot.get(1,2);
	double r20 = rot.get(2,0),r21 = rot.get(2,1),r22 = rot.get(2,2);
	double tr0 = trans.get(0,0),tr1 = trans.get(1,0),tr2 = trans.get(2,0);
	StringWriter ret = new StringWriter();
	PrintWriter  pw  = new PrintWriter(ret);
	for (Pair p = fPair_;p != null;p = p.next()) {
	    if (!p.isOfInterest()) continue;
	    
	    Assembly ass1 = (Assembly)p.getObject1();
	    Assembly ass2 = (Assembly)p.getObject2();
	    Atom a1 = ass1.getMainAtom();
	    Atom a2 = ass2.getMainAtom();
	    if (a1 == null || a2 == null) continue;
	    double x2 = a2.getDerivedX();
	    double y2 = a2.getDerivedY();
	    double z2 = a2.getDerivedZ();
	    
	    double x_displ = r00*x2 + r01*y2 + r02*z2 +	tr0 - a1.getDerivedX();
	    double y_displ = r10*x2 + r11*y2 + r12*z2 +	tr1 - a1.getDerivedY();
	    double z_displ = r20*x2 + r21*y2 + r22*z2 +	tr2 - a1.getDerivedZ();
	    pw.format("%8.4f ",x_displ);
	    pw.format("%8.4f ",y_displ);
	    pw.format("%8.4f ",z_displ);
	}
	pw.println();
	pw.flush();
	pw.close();
	return ret.getBuffer();
    } 
   
    /**
     * Returns string with values of displacement for each CA atom. The 
     * diplacement is a vector between initial and interpolated 
     * by angle structures. Amount of interpolation is controlled by parameter
     * fraction.
     *
     * @param staticGroup id of static (not moving) block.
     * @param fraction value of interpolation [0,1].
     *
     * @return string with values of displacement for each CA atom.
     */
    public StringBuffer print(int staticGroup,double fraction)
    {
	StringWriter ret = new StringWriter();
	if (rotations_ == null || translations_ == null)
	    return ret.getBuffer();

	if (fraction < 0 || fraction > 1) {
	    System.err.println("Interpolation fraction should be between " +
			       "0 and 1.");
	    return ret.getBuffer();
	}
	
	setRelativeIndex(staticGroup);
	
	// Getting global transformation
	Matrix g_rot   = rotations_[0];
	Matrix g_trans = translations_[0];
	if (g_rot == null || g_trans == null) {
	    System.err.println("No global tranformation found.");
	    return ret.getBuffer();
	}

	int n_rigids = rotations_.length - 1;
	// Array with fractional transformation for each rigid block
	Matrix[] f_rotations    = new Matrix[n_rigids + 1];
	Matrix[] f_translations = new Matrix[n_rigids + 1];
	f_rotations[0]    = g_rot;
	f_translations[0] = g_trans;

	for (int r = 1;r <= n_rigids;r++) {

	    int relative = relative_index_[r];
	    Matrix r_rot   = rotations_[relative];
	    Matrix r_trans = translations_[relative];
	    if (r_rot == null || r_trans == null) {
		System.err.println("No relative transformation found.");
		return ret.getBuffer();
	    }

	    Matrix l_rot   = rotations_[r];
	    Matrix l_trans = translations_[r];
	    if (l_rot == null || l_trans == null) {
		System.err.println("No local transformation found.");
		return ret.getBuffer();
	    }

 	    // Rotation and translation from relative to local superpositions
	    Matrix rot   = l_rot.times(r_rot.transpose());
	    Matrix trans = l_trans.minus(rot.times(r_trans));

	    // Calculating rotation axis and angle
	    EigenvalueDecomposition eig = rot.eig();
 	    double[] eigen_values  = eig.getRealEigenvalues();
 	    Matrix   eigen_vectors = eig.getV();
	    int index = 0;
	    double delta = Math.abs(1 - eigen_values[0]);
	    for (int i = 1;i < eigen_values.length;i++) {
		double new_delta = Math.abs(1 - eigen_values[i]);
		if (new_delta < delta) {
		    index = i;
		    delta = new_delta;
		}
	    }
	    Matrix axis =
		eigen_vectors.getMatrix(0,eigen_vectors.getRowDimension() - 1,
					index,index);
	    double cos = eigen_values[0];
	    if (index == 0) cos = eigen_values[1];

	    // Determining angle sign
	    double sin  = Math.sqrt(1 - cos*cos);
	    double tmp  = axis.get(0,0)*axis.get(1,0)*(1 - cos);
	    double v1   = tmp - axis.get(2,0)*sin;
	    double v2   = tmp + axis.get(2,0)*sin;
	    double v_xy = rot.get(0,1);
	    if (Math.abs(v_xy - v1) > Math.abs(v_xy - v2)) sin = -sin;

	    // Calculating angle
	    double angle = Math.asin(sin);
	    //System.out.println("angle = " + angle);

	    // Local interpolation. Note minus sign in front of fractions.
	    // The sign reflect the fact that we calculated fractional
	    // rotation matrix and translation vector for first molecule.
	    double f_angle  = -fraction*angle;
	    Matrix fl_rot   = new Matrix(calcRotationMatrix(axis,f_angle));
	    Matrix fl_trans = trans.times(-fraction);

	    // Relative interpolation
	    double[][] I = {{1,0,0},{0,1,0},{0,0,1}};
	    Matrix fr_rot   = new Matrix(I);
	    Matrix fr_trans = new Matrix(fl_trans.getRowDimension(),
					 fl_trans.getColumnDimension());
	    if (relative > 0) { // Relative to other rigid block
		fr_rot   = f_rotations[relative];
		fr_trans = f_translations[relative];
	    }

	    // Combining interpolations
	    f_rotations[r]    = fl_rot.times(fr_rot);
	    f_translations[r] = fl_rot.times(fr_trans).plus(fl_trans);
// 	    double[][] tmp1 = fl_rot.getArray();
// 	    double[][] tmp2 = f_rotations[r].getArray();
// 	    for (int i1 = 0;i1 < 3;i1++) {
// 		for (int i2 = 0;i2 < 3;i2++) {
// 		    System.out.format("%6.2f ",tmp1[i1][i2]);
// 		    System.out.format("%6.2f ",tmp2[i1][i2]);
// 		}
// 		System.out.println();
// 	    }
// 	    System.out.println();
// 	    for (int i = 0;i < 3;i++)
// 		System.out.format("%6.2f ",l_trans.get(i,0) -
// 				  f_translations[r].get(i,0));
// 	    System.out.println();
// 	    System.out.println();
	}

	return print(f_rotations,f_translations,fraction);
    }

    // Printing (infinitesimal) displacement from interpolation 
    private StringBuffer print(Matrix[] rs,Matrix ts[],double fraction)
    {
	StringWriter ret = new StringWriter();
	PrintWriter  pw  = new PrintWriter(ret);
	for (Pair p = fPair_;p != null;p = p.next()) {
	    if (!p.isOfInterest()) continue;
	    Assembly ass1 = (Assembly)p.getObject1();
	    Assembly ass2 = (Assembly)p.getObject2();
	    Atom a1 = ass1.getMainAtom();
	    Atom a2 = ass2.getMainAtom();
	    if (a1 == null || a2 == null) continue;
	    double x = a1.getDerivedX();
	    double y = a1.getDerivedY();
	    double z = a1.getDerivedZ();
	    double x_displ, y_displ, z_displ;
	    int r = p.getIntValue();
	    if (r <= 0 || rs[r] == null || ts[r] == null) {
		Matrix rot = rs[0], trans = ts[0];
		double x2 = a2.getDerivedX();
		double y2 = a2.getDerivedY();
		double z2 = a2.getDerivedZ();
		x_displ = rot.get(0,0)*x2 + rot.get(0,1)*y2 + rot.get(0,2)*z2 +
		    trans.get(0,0) - x;
		y_displ = rot.get(1,0)*x2 + rot.get(1,1)*y2 + rot.get(1,2)*z2 +
		    trans.get(1,0) - y;
		z_displ = rot.get(2,0)*x2 + rot.get(2,1)*y2 + rot.get(2,2)*z2 +
		    trans.get(2,0) - z;
		x_displ *= fraction;
		y_displ *= fraction;
		z_displ *= fraction;
	    } else {
		Matrix rot = rs[r], trans = ts[r];
		x_displ = rot.get(0,0)*x + rot.get(0,1)*y + rot.get(0,2)*z +
		    trans.get(0,0) - x;
		y_displ = rot.get(1,0)*x + rot.get(1,1)*y + rot.get(1,2)*z +
		    trans.get(1,0) - y;
		z_displ = rot.get(2,0)*x + rot.get(2,1)*y + rot.get(2,2)*z +
		    trans.get(2,0) - z;
	    }
	    pw.format("%8.4f ",x_displ);
	    pw.format("%8.4f ",y_displ);
	    pw.format("%8.4f ",z_displ);
	}
	pw.println();
	pw.flush();
	pw.close();
	return ret.getBuffer();
    }

    private double[][] calcRotationMatrix(Matrix vector,double angle)
    {
	double x = vector.get(0,0), y = vector.get(1,0), z = vector.get(2,0);
	double cos = Math.cos(angle), sin = Math.sin(angle), C = 1 - cos;
	double xs  = x*sin, ys  = y*sin, zs  = z*sin;
	double xC  = x*C,   yC  = y*C,   zC  = z*C;
	double xyC = x*yC,  yzC = y*zC,  zxC = z*xC;
	double[][] ret = {{x*xC + cos, xyC  - zs,  zxC + ys},
			  {xyC  + zs,  y*yC + cos, yzC - xs},
			  {zxC  - ys,  yzC  + xs,  z*zC + cos}};
	return ret;
    }

    private void setRelativeIndex(int staticGroup)
    {
    }
}